# -*- coding: utf-8 -*-

import json

server_error_codes = {
    "unknown": 400,
    "missing_object": 404
}

try:
    from werkzeug.exceptions import HTTPException
except:
    # No need to bind objects here to dependency-sink, as the user
    # will be notified when he tries to use Slicer or run_server about
    # the missing package
    HTTPException = object


class ServerError(HTTPException):
    code = 500
    error_type = "default"
    def __init__(self, message=None, exception=None, **details):
        super(ServerError, self).__init__()
        self.message = message
        self.exception = exception
        self.details = details
        self.help = None

    def get_body(self, environ):
        error = {
            "message": self.message,
            "type": self.__class__.error_type
        }

        if self.exception:
            error["reason"] = str(self.exception)

        if self.details:
            error.update(self.details)

        string = json.dumps({"error": error}, indent=4)
        return string

    def get_headers(self, environ):
        """Get a list of headers."""
        return [('Content-Type', 'application/json')]


class RequestError(ServerError):
    error_type = "request"
    code = 400


class NotAuthorizedError(ServerError):
    code = 403
    error_type = "not_authorized"


class NotAuthenticatedError(ServerError):
    code = 401
    error_type = "not_authenticated"

    def __init__(self, message=None, exception=None, realm=None, **details):
        super(NotAuthenticatedError, self).__init__(message,
                                                    exception,
                                                    **details)
        self.message = message
        self.exception = exception
        self.details = details
        self.help = None
        self.realm = realm or "Default"

    def get_headers(self, environ):
        """Get a list of headers."""
        headers = super(NotAuthenticatedError, self).get_headers(environ)
        headers.append(('WWW-Authenticate', 'Basic realm="%s"' % self.realm))
        return headers


# TODO: Rename this to plain NotFoundError
class PageNotFoundError(ServerError):
    code = 404
    error_type = "not_found"
    def __init__(self, message=None):
        super(PageNotFoundError, self).__init__(message)


# TODO: Rename this to ObjectNotFoundError
class NotFoundError(ServerError):
    code = 404
    error_type = "object_not_found"
    def __init__(self, obj, objtype=None, message=None):
        super(NotFoundError, self).__init__(message)
        self.details = { "object": obj }

        if objtype:
            self.details["object_type"] = objtype

        if not message:
            self.message = "Object '%s' of type '%s' was not found" % (obj, objtype)
        else:
            self.message = message





# -*- encoding: utf-8 -*-
"""Pytho compatibility utilities"""

from __future__ import absolute_import

import sys

py3k = sys.version_info >= (3, 0)

if py3k:
    string_type = str
    binary_type = bytes
    text_type = str
    int_types = int,
    iterbytes = iter

    from urllib.parse import urlparse
    from urllib.request import urlopen, build_opener
    from urllib.request import HTTPPasswordMgrWithDefaultRealm
    from urllib.request import HTTPBasicAuthHandler
    from urllib.parse import urlencode
    from configparser import ConfigParser
    from io import StringIO
    from queue import Queue
    from functools import reduce

    def to_unicode(s):
        return str(s)

    def to_str(b):
        return b.decode("utf-8")

    def open_unicode(filename):
        return open(filename, encoding="utf-8")

else:
    string_type = basestring
    binary_type = str
    text_type = unicode
    int_types = int, long

    from urlparse import urlparse
    from urllib2 import urlopen, build_opener
    from urllib2 import HTTPPasswordMgrWithDefaultRealm
    from urllib2 import HTTPBasicAuthHandler
    from urllib import urlencode
    from ConfigParser import SafeConfigParser as ConfigParser
    from StringIO import StringIO
    from Queue import Queue
    reduce = reduce

    def to_str(b):
        return b

    def to_unicode(s):
        if isinstance(s, unicode):
            return s
        s = str(s)
        for enc in ('utf8', 'latin-1'):
            try:
                return unicode(s, enc)
            except UnicodeDecodeError:
                pass

        raise ValueError("Cannot decode for unicode using any of the default "
                         "encodings: %s" % s)

    def open_unicode(filename):
        return open(filename)



# -*- encoding: utf-8 -*-

from __future__ import absolute_import

from flask import Request, Response, request, g

import codecs
import json
import csv

from .errors import *
from ..formatters import csv_generator, JSONLinesGenerator, SlicerJSONEncoder
from .. import compat


def str_to_bool(string):
    """Convert a `string` to bool value. Returns ``True`` if `string` is
    one of ``["true", "yes", "1", "on"]``, returns ``False`` if `string` is
    one of  ``["false", "no", "0", "off"]``, otherwise returns ``None``."""

    if string is not None:
        if string.lower() in ["true", "yes", "1", "on"]:
            return True
        elif string.lower() in["false", "no", "0", "off"]:
            return False

    return None


def validated_parameter(args, name, values=None, default=None,
                        case_sensitive=False):
    """Return validated parameter `param` that has to be from the list of
    `values` if provided."""

    param = args.get(name)
    if param:
        param = param.lower()
    else:
        param = default

    if not values:
        return param
    else:
        if values and param not in values:
            list_str = ", ".join(values)
            raise RequestError("Parameter '%s' should be one of: %s"
                            % (name, list_str) )
        return param


class CustomDict(dict):
    def __getattr__(self, attr):
        try:
            return super(CustomDict, self).__getitem__(attr)
        except KeyError:
            return super(CustomDict, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        self.__setitem__(attr,value)


# Utils
# =====

def jsonify(obj):
    """Returns a ``application/json`` `Response` object with `obj` converted
    to JSON."""

    if g.prettyprint:
        indent = 4
    else:
        indent = None

    encoder = SlicerJSONEncoder(indent=indent)
    encoder.iterator_limit = g.json_record_limit
    data = encoder.iterencode(obj)

    return Response(data, mimetype='application/json')


def formatted_response(response, fields, labels, iterable=None):
    """Wraps request which returns response that can be formatted. The
    `data_attribute` is name of data attribute or key in the response that
    contains formateable data."""

    output_format = validated_parameter(request.args, "format",
                                        values=["json", "json_lines", "csv"],
                                        default="json")

    header_type = validated_parameter(request.args, "header",
                                      values=["names", "labels", "none"],
                                      default="labels")

    # Construct the header
    if header_type == "names":
        header = fields
    elif header_type == "labels":
        header = labels
    else:
        header = None


    # If no iterable is provided, we assume the response to be iterable
    iterable = iterable or response

    if output_format == "json":
        return jsonify(response)
    elif output_format == "json_lines":
        return Response(JSONLinesGenerator(iterable),
                        mimetype='application/x-json-lines')
    elif output_format == "csv":
        generator = csv_generator(iterable,
                                 fields,
                                 include_header=bool(header),
                                 header=header)

        headers = {"Content-Disposition": 'attachment; filename="facts.csv"'}

        return Response(generator,
                        mimetype='text/csv',
                        headers=headers)





# -*- coding: utf-8 -*-

__all__ = (
    "Store"
)


# Note: this class does not have much use right now besides being discoverable
# by custom plugin system in cubes.
# TODO: remove requirement for store_name and store_type
class Store(object):
    """Abstract class to find other stores through the class hierarchy."""

    """Name of a model provider type associated with this store."""
    related_model_provider = None
    store_type = None

    def __init__(self, **options):
        # We store the parsed options from the store configuration here
        self.options = options

        # TODO: this is just backward compatibility, remove this (make this
        # class variable)
        self.store_type = options.get("store_type")



# -*- coding: utf-8 -*-
from flask import current_app
from werkzeug.local import LocalProxy

# Application Context
# ===================
#
# Readability proxies

def _get_workspace():
    return current_app.cubes_workspace

def _get_logger():
    return current_app.cubes_workspace.logger

workspace = LocalProxy(_get_workspace)
logger = LocalProxy(_get_logger)

